"""Top-level package for MoiToi Docker Hive."""

__author__ = """Andres Kepler"""
__email__ = 'andres@kepler.ee'
__version__ = '0.1.8'

import logging.config
import uuid
from os.path import expanduser

DEFAULT_CONFIG_DIR = expanduser("~") + "/.mdh"
DEFAULT_CONFIG_FILE = DEFAULT_CONFIG_DIR + "/mdh.yaml"
DEFAULT_MDH_PASSWORD = "Change;Me"
DEFAULT_MDH_CLUSTER_ID = "mdh" + str(uuid.uuid1()).replace("-", "")
DEFAULT_MDH_LD_PORT = 5000
DEFAULT_MDH_BACKEND_PORT = 5000

logging.config.dictConfig({
    'version': 1,
    'formatters': {
        'default': {
            'format': '%(asctime)s [%(levelname)s] %(name)s:%(lineno)d %(funcName)s - %(message)s'
        },
    },
    'handlers': {
        'log': {
            'class': 'logging.StreamHandler',
            'formatter': 'default',
        },
    },
    'loggers': {
        '': {
            'handlers': ['log'],
            'level': 'INFO',
        }
    }
})

DEFAULT_CONFIG = [{
    "provider": {
        "name": "local",
        "dockerhost": "localhost"
    },
},
    {"password": DEFAULT_MDH_PASSWORD},
    {"backend_port": DEFAULT_MDH_BACKEND_PORT}

]



"""Console script for moitoi_docker_hive."""
import argparse
import os
import sys

import yaml

from moitoi_docker_hive import DEFAULT_CONFIG_DIR, DEFAULT_MDH_PASSWORD, \
    DEFAULT_MDH_CLUSTER_ID, DEFAULT_CONFIG, DEFAULT_CONFIG_FILE
from .main import ClusterHandler


def parse_args():
    """Console script for moitoi_docker_hive."""
    parser = argparse.ArgumentParser()
    parser.add_argument('cmd', nargs='*', choices=["init", "show", "+", "-"], help='Cluster commands')
    parser.add_argument('--private-key', default="{}/id_rsa".format(DEFAULT_CONFIG_DIR),
                        help='ssh private key location')
    parser.add_argument('--public-key', default="{}/id_rsa.pub".format(DEFAULT_CONFIG_DIR),
                        help='ssh public key location')
    parser.add_argument('--hive-password', default=os.environ.get("MDH_PASSWORD", DEFAULT_MDH_PASSWORD),
                        help='hive password for keys/user')
    parser.add_argument('--cluster-id', default=os.environ.get("MDH_CLUSTER_ID", DEFAULT_MDH_CLUSTER_ID),
                        help='Uniq cluster id')
    parser.add_argument('--node-name', help='Docker container on append/remove to cluster')
    return parser.parse_args()


def prepare_env():
    if not os.path.isdir(DEFAULT_CONFIG_DIR):
        os.mkdir(DEFAULT_CONFIG_DIR, 0o750)

    if not os.path.isfile(DEFAULT_CONFIG_FILE):
        with open(DEFAULT_CONFIG_FILE, 'w') as file:
            yaml.dump(DEFAULT_CONFIG, file)

    with open(DEFAULT_CONFIG_FILE) as file:
        documents = yaml.full_load(file)[0]
        for item, doc in documents.items():
            if isinstance(doc, dict):
                for _i, _d in doc.items():
                    os.environ["MDH_{}_{}".format(item.upper(), _i.upper())] = _d
            else:
                os.environ["MDH_{}".format(item.upper())] = doc


def worker(args):
    cli = ClusterHandler(args)
    cli.load_ssh_keys()
    cli.invoke()


def main():
    prepare_env()
    worker(parse_args())
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



from flask_wtf import FlaskForm
from wtforms import IntegerField
from wtforms.validators import DataRequired


class ClusterSetupForm(FlaskForm):
    clusters = IntegerField('Clusters', validators=[DataRequired()])
    replicas = IntegerField('Replicas', validators=[DataRequired()])



import os


class BaseConfig(object):
    """
     This is a great place to pass configuration parameters from the
     Environment into your application
    """
    # DB_NAME = os.environ['DB_NAME']
    SECRET_KEY = os.environ.get('SECRET_KEY', '78w0o5tuuGex5Ktk8VvVDF9Pw3jv1MVE')
    pass



"""
MDH REST Api module
"""
import os
import re

from flask import Flask
from flask_restful import Api

from rest.controller import ServiceAPI, StateAPI, Root

app = Flask(__name__)
app.config['DEBUG'] = re.match("(1|TRUE|Y|True)", os.environ.get('DEBUG', "false"))
api = Api(app)

api.add_resource(ServiceAPI, '/Service')
api.add_resource(StateAPI, '/State')
api.add_resource(Root, '/')

if __name__ == '__main__':
    app.run(debug=True, host='0.0.0.0', port=5000)



import uuid


class Args:
    def __init__(self, cluster_id=None, node_name=None, cmd=None, private_key=None, public_key=None,
                 hive_password=None):
        self.cluster_id = cluster_id or "mdh" + str(uuid.uuid1()).replace("-", "")
        self.node_name = node_name
        self.cmd = cmd
        self.private_key = private_key
        self.public_key = public_key
        self.hive_password = hive_password or self.cluster_id



import os
import logging
import re

from flask import Flask, render_template, redirect
from flask_bootstrap import Bootstrap
from flask_nav import Nav
from flask_nav.elements import Navbar, View, Subgroup, Link, Separator
from config import BaseConfig
from forms import ClusterSetupForm

cluster_id = os.environ.get("MDH_CLUSTER_ID")

nav = Nav()

# registers the "top" menubar
nav.register_element('frontend_top', Navbar(
    View('MDH', 'index'),
    Subgroup(
        'Service',
        Link('Add service', 'Service?cmd=add_service'.format(cluster_id)),
        Separator(),
        Link('Add node', 'Service?cmd=add_node&cluster_id={}'.format(cluster_id)),
        Link('Remove node', 'Service?cmd=rm_node&cluster_id={}'.format(cluster_id)),
        Separator(),
        Link('Cluster state', 'State?cluster_id={}'.format(cluster_id)),
        Link('Clusters state', 'State'),
        Separator(),
        Link('Grafana', 'Grafana'),

    ),
))

app = Flask(__name__)
app.config['DEBUG'] = re.match("(1|TRUE|Y|True)", os.environ.get('DEBUG', "false"))
nav.init_app(app)

if __name__ != '__main__':
    gunicorn_logger = logging.getLogger('gunicorn.error')
    app.logger.handlers = gunicorn_logger.handlers
    app.logger.setLevel(gunicorn_logger.level)

app.config.from_object(BaseConfig)
Bootstrap(app)


@app.route('/', methods=('GET', 'POST'))
def index():
    form = ClusterSetupForm()
    if form.validate_on_submit():
        return redirect('/')
    return render_template('index.html', hostname=os.environ.get("MDH_HOSTNAME"), mdh_cluster_id=cluster_id, form=form)


if __name__ == '__main__':
    app.run(host='0.0.0.0', port=5000, debug=True)



"""Main module."""
import logging
import os
from argparse import ArgumentParser

from paramiko import RSAKey
import docker

DEFAULT_MDH_LD_PORT = 5000
DEFAULT_MDH_BACKEND_PORT = 5000


class ClusterHandler:
    ssh_public_key = None
    cluster_id = None
    cluster_network = None
    cluster_data_volume = None
    lb_container = None
    elsa_container = None
    prom_container = None
    grafana_container = None
    mdh_backends = []

    def __init__(self, args):
        self.args = args
        self.cluster_id = args.cluster_id
        self.docker_client = docker.from_env()
        self.prepare()

    @staticmethod
    def generate_ssh_key_pairs(filename, phrase):
        """ SSH key pairs generator for remote docker host provider

        Args:
            filename (str):The filename path where ssh keys will be saved.
            phrase (str):The password phrase.

        Returns:

        """
        # generating private key
        prv = RSAKey.generate(bits=2048, progress_func=None)
        prv.write_private_key_file(filename, password=phrase)

        # generating public key
        pub = RSAKey(filename=filename, password=phrase)
        with open("%s.pub" % filename, "w") as f:
            f.write("%s %s" % (pub.get_name(), pub.get_base64()))
            f.write(" %s" % "moitoi_docker_hive_key")

    def load_ssh_keys(self):
        """ Helper method to load ssh keys
        """
        if not os.path.isfile(self.args.private_key):
            self.generate_ssh_key_pairs(self.args.private_key, self.args.hive_password)

        if not self.ssh_public_key:
            with open(self.args.public_key) as f:
                self.ssh_public_key = ''.join(f.readline())

    def _lookup_container_by_label(self, label_key):
        """ Lookup from current cluster container by label key

        Args:
            label_key (str):The key of label.

        Returns:
            containers (list): The list of found container objects.

        """
        containers = []
        for container in self.docker_client.containers.list():
            if container.labels.get(label_key) and container.labels.get("mdh_cluster_id") == self.cluster_id:
                containers.append(container)
        return containers

    def _lookup_container_by_name(self, container_name):
        """ Lookup from current cluster container by container name

        Args:
            container_name (str):The container name to lookup.

        Returns:
            containers (list): The list of found container objects.

        """
        containers = []
        for container in self.docker_client.containers.list():
            if container.name == container_name and container.labels.get("mdh_cluster_id") == self.cluster_id:
                containers.append(container)
        return containers

    def _lookup_container_by_cluster_id(self):
        """ List all current cluster containers

        Returns:
            containers (list): The list of found container objects.

        """
        containers = []
        for container in self.docker_client.containers.list():
            if container.labels.get("mdh_cluster_id") == self.cluster_id:
                containers.append(container)
        return containers

    def _show_all_mdh_containers(self):
        """ Helper method to list all mdh containers
        Returns:
            containers (list): The list of found container objects.

        """
        return list(filter(lambda x: x.labels.get("mdh_cluster_id"), self.docker_client.containers.list()))

    def _show_all_mdh_cluster_ids(self):
        """ Helper method to list all mdh cluster id
        Returns:
            res (list): The list of found cluster id's.

        """
        containers = self.docker_client.containers.list(filters={"label": ["mdh_lb=true"]})
        res = []
        for container in containers:
            cluster_id = container.labels.get("mdh_cluster_id")
            if cluster_id not in res:
                res.append(cluster_id)
        return res

    def _get_or_create_data_volume(self):
        """ Helper method to get or crate data volume for image
        Returns:
            mdh_volumes (list): The list of found or created volume objects.

        """
        mdh_volumes = []
        for _v in self.docker_client.volumes.list():
            labels = _v.attrs['Labels']
            if not labels:
                continue
            if labels.get("mdh_data_volume") and labels.get("mdh_cluster_id", "NULL") == self.cluster_id:
                mdh_volumes.append(_v)
        if not len(mdh_volumes):
            volume = self.docker_client.volumes.create(name='{}_data'.format(self.cluster_id), driver='local',
                                                       labels={"mdh_data_volume": "true",
                                                               "mdh_cluster_id": self.cluster_id})
            mdh_volumes.append(volume)
        return mdh_volumes

    def _get_or_create_network(self):
        """ Helper method to get or create network for service
        Returns:
            mdh_networks (list): The list of found or created network objects.

        """
        mdh_networks = []
        for _n in self.docker_client.networks.list():
            labels = _n.attrs['Labels']
            if not labels:
                continue
            if labels.get("mdh_network") and labels.get("mdh_cluster_id", "NULL") == self.cluster_id:
                mdh_networks.append(_n)
                return mdh_networks

        if not len(mdh_networks):
            network = self.docker_client.networks.create(name='{}_network'.format(self.cluster_id), driver='bridge',
                                                         labels={"mdh_network": "true",
                                                                 "mdh_cluster_id": self.cluster_id})
            mdh_networks.append(network)
        return mdh_networks

    def _get_or_create_image(self, image_type):
        """ Helper method to get or create docker image
        Returns:
            mdh_images (list): The list of found or created docker image objects.

        """
        mdh_images = []
        for _i in self.docker_client.images.list():
            labels = _i.labels
            if not labels:
                continue
            if labels.get("mdh_{}".format(image_type), "NULL") == "true":
                logging.info(_i)
                mdh_images.append(_i)
        if not len(mdh_images):
            labels = {}
            labels["mdh_{}".format(image_type)] = "true"
            image = self.docker_client.images.build(
                path=os.path.dirname(os.path.realpath(__file__)) + "/assets/{}".format(image_type),
                tag="mdh_{}".format(image_type),
                pull=True,
                labels=labels
            )
            mdh_images.append(image)

        return mdh_images

    def _lookup_next_lb_port(self):
        """ Helper method to identify next available host port
        Returns:
            containers (int): Next next available host port.

        """
        p = []
        for container in self.docker_client.containers.list():
            if container.labels.get("mdh_lb", "null") == "true":
                for _k, _v in container.attrs.get('NetworkSettings').get("Ports").items():
                    if _v:
                        for __v in _v:
                            p.append(int(__v.get('HostPort')))
        if p:
            return sorted(p)[-1] + 1
        return DEFAULT_MDH_LD_PORT

    def setup_cluster_infra(self):
        """ Init cluster infra.
        Returns:
            cluster_data (dict): Created cluster data.

        """
        logging.info("Creating new cluster with id {}".format(self.cluster_id))

        # Build images
        images_to_build = ["statsd", "lb", "elsa", "backend", "prom", "grafana", ]
        images = {}
        for image in images_to_build:
            logging.info("Building {} image".format(image.upper()))
            images["{}_image"] = self._get_or_create_image(image_type=image)

        # Create volumes
        _data_volume = {self._get_or_create_data_volume()[0].attrs['Name']: {'bind': '/data', 'mode': 'rw'}}
        _docker_socket_volume = {'/var/run/docker.sock': {'bind': '/var/run/docker.sock', 'mode': 'rw'}}

        # Create network
        _networks = self._get_or_create_network()
        self.cluster_network = _networks[0]
        logging.info("Network {}".format(self.cluster_network.short_id))

        # Run LB container
        next_available_port, mdh_lb = self.run_lb_container(_data_volume)

        # Run monitoring stuff first
        for name in ["statsd", "prom", "grafana"]:
            self.run_container(name)

        # Run backend container
        self.run_backend_container(_data_volume)

        # Run ELSA container
        self.run_elsa_container(_docker_socket_volume)

        # Start LB container if its not running
        if self.lb_container.status != "running":
            self.lb_container.start()

        res = {"mdh_cluster_id": self.cluster_id,
               "mdh_backends": self.mdh_backends[0].short_id,
               "mdh_lb": self.lb_container.short_id,
               "mdh_elsa": self.elsa_container.short_id,
               "mdh_prom": self.prom_container.short_id,
               "mdh_statsd": self.statsd_container.short_id,
               "mdh_grafana": self.grafana_container.short_id,
               "mdh_network": self.cluster_network.short_id,
               "mdh_port": next_available_port}
        logging.info(res)
        return res

    def run_backend_container(self, _data_volume):
        """ Run backend container

        Args:
            _data_volume (dict): Shared volume

        Returns (obj): Container object

        """
        self.mdh_backends = self._lookup_container_by_label("mdh_backend")
        if not len(self.mdh_backends):
            container = self.run_mdh_backend(_data_volume, "1")
            logging.info("Running BACKEND conatiner {}, Status: {}".format(container.short_id,
                                                                           container.status))
            self.mdh_backends.append(container)
        return self.mdh_backends

    def run_lb_container(self, _data_volume):
        """Run load balancer container

        Args:
            _data_volume (dict): Shared volume

        Returns (tuple): Tuple of attached load balancer external port and load balancer Container object

        """
        next_available_port = DEFAULT_MDH_LD_PORT
        mdh_lb = self._lookup_container_by_label("mdh_lb")
        if not len(mdh_lb):
            next_available_port = self._lookup_next_lb_port()
            self.lb_container = self.docker_client.containers.run('mdh_lb',
                                                                  name="mdh_lb_{}".format(self.cluster_id),
                                                                  labels=dict(mdh_lb="true",
                                                                              mdh_cluster_id=self.cluster_id),
                                                                  ports={'5000/tcp': next_available_port},
                                                                  environment={"MDH_CLUSTER_ID": self.cluster_id,
                                                                               "MDH_BACKEND_PORT": os.environ.get(
                                                                                   "MDH_BACKEND_PORT", 5000),
                                                                               "MDH_HOSTNAME": "mdh_ld"
                                                                               },
                                                                  volumes=_data_volume,
                                                                  detach=True)
            logging.info("Running LB conatiner {}, Status: {}".format(self.lb_container.short_id,
                                                                      self.lb_container.status))
            self.cluster_network.connect(self.lb_container, aliases=["mdh_lb"])
            mdh_lb.append(self.lb_container.short_id)
        return next_available_port, mdh_lb

    def run_container(self, container_name):
        """ Run general container. If vontainer already exist it just returns list with container
        Args:
            container_name (str): Container image name for example prom, grafana eg

        Returns (list): Container list

        """
        container_list = self._lookup_container_by_label("mdh_{}".format(container_name))
        if not len(container_list):
            env = {"MDH_CLUSTER_ID": self.cluster_id, "MDH_HOSTNAME": "mdh_{}".format(container_name)}
            if container_name == "grafana":
                env.update({"GF_SECURITY_ADMIN_PASSWORD": self.cluster_id[:10]})
            container = self.docker_client.containers.run('mdh_{}'.format(container_name),
                                                          name="mdh_{}_{}".format(container_name, self.cluster_id),
                                                          labels={"mdh_{}".format(container_name): "true",
                                                                  "mdh_cluster_id": self.cluster_id},
                                                          environment=env,
                                                          volumes_from=[self.lb_container.name],
                                                          detach=True)
            self.cluster_network.connect(container, aliases=["mdh_{}".format(container_name)])
            setattr(self, "{}_container".format(container_name), container)
            container_list.append(container.short_id)
            logging.info("Running {} conatiner {}, Status: {}".format(container_name.upper(),
                                                                      container.short_id,
                                                                      container.status))
        return container_list

    def run_elsa_container(self, _docker_socket_volume):
        """ Run ELSA container.
        Args:
            _docker_socket_volume (dict): Shared host volume to access docker info

        Returns (list): Container list

        """
        mdh_elsa = self._lookup_container_by_label("mdh_elsa")
        if not len(mdh_elsa):
            elsa_volumes = {}
            elsa_volumes.update(_docker_socket_volume)
            self.elsa_container = self.docker_client.containers.run('mdh_elsa',
                                                                    name="mdh_elsa_{}".format(self.cluster_id),
                                                                    labels=dict(mdh_elsa="true",
                                                                                mdh_cluster_id=self.cluster_id),
                                                                    environment={"MDH_CLUSTER_ID": self.cluster_id,
                                                                                 "MDH_BACKEND_PORT": os.environ.get(
                                                                                     "MDH_BACKEND_PORT", 5000),
                                                                                 "MDH_HOSTNAME": "mdh_elsa"
                                                                                 },
                                                                    volumes_from=[self.lb_container.name],
                                                                    volumes=elsa_volumes,
                                                                    detach=True)
            self.cluster_network.connect(self.elsa_container, aliases=["mdh_elsa"])
            mdh_elsa.append(self.elsa_container.short_id)
            logging.info("Running ELSA conatiner {}, Status: {}".format(self.elsa_container.short_id,
                                                                        self.elsa_container.status
                                                                        ))
        return mdh_elsa

    def run_mdh_backend(self, _data_volume, nr):
        """ Run mdh_backend image

        Args:
            _data_volume (dict): Shared data volume
            nr (str): Backend container number

        Returns:
            container (object): Created container object

        """
        container = self.docker_client.containers.run('mdh_backend',
                                                      name="mdh_backend{}_{}".format(nr, self.cluster_id),
                                                      labels=dict(mdh_backend="true",
                                                                  mdh_cluster_id=self.cluster_id,
                                                                  mdh_backend_member_nr=nr),
                                                      environment={"MDH_CLUSTER_ID": self.cluster_id,
                                                                   "MDH_BACKEND_PORT":
                                                                       os.environ.get("MDH_BACKEND_PORT", 5000),
                                                                   "MDH_HOSTNAME": "mdh_backend{}".format(nr)
                                                                   },
                                                      volumes=_data_volume,
                                                      detach=True)
        self.cluster_network.connect(container, aliases=["mdh_backend{}".format(nr)])
        return container

    def show_cluster(self, show_all=False):
        """ Show cluster status

        Args:
            show_all (bool): True returns all clusters, False return only self.cluster_id info

        Returns:
            res (list): List of cluster(s)

        """
        res = []
        clusters = self._show_all_mdh_cluster_ids()
        if isinstance(self.args, ArgumentParser):
            print("Cluster ID\t\t\t\tContainer Name\t\t\t\t\t\t\tContainer ID")
        if self.cluster_id not in clusters or show_all:
            for cluster_id in clusters:
                containers = []
                for container in list(filter(lambda x: x.get("Labels").get("mdh_cluster_id", "not_set") == cluster_id,
                                             self.docker_client.api.containers())):
                    containers.append(container)
                    if isinstance(self.args, ArgumentParser):
                        print("{}\t{}\t{}".format(container.get("Labels").get("mdh_cluster_id"),
                                                  container.get("Names")[0].split("/")[1],
                                                  container.get("Id")))
                res.append({cluster_id: containers})

        else:
            for container in list(filter(lambda x: x.get("Labels").get("mdh_cluster_id", "not_set") == self.cluster_id,
                                         self.docker_client.api.containers())):
                if isinstance(self.args, ArgumentParser):
                    print("{}\t{}\t{}".format(container.get("Labels").get("mdh_cluster_id"),
                                              container.get("Names")[0].split("/")[1], container.get("Id")))
                res.append(container)
        return res

    def get_all_backend_nodes(self):
        """ Get cluster backned conatiners

        Returns (list): List of sorted backend containers

        """
        containers = self._lookup_container_by_label("mdh_backend")
        return sorted(containers, key=lambda x: x.labels.get("mdh_backend_member_nr"))

    def add_node_to_cluster(self):
        """ Add backend node to cluster

        Returns:
            res (list): Cluster status

        """
        self.mdh_backends = self.get_all_backend_nodes()
        last_backend_nr = sorted(list(map(lambda x: int(x.labels.get("mdh_backend_member_nr")), self.mdh_backends)))[-1]
        self.run_mdh_backend(self.cluster_data_volume, str(last_backend_nr + 1))
        return self.show_cluster()

    def remove_node_from_cluster(self):
        """ Removes node from cluster

        Returns:
            res (list): Cluster status

        """
        self.mdh_backends = self.get_all_backend_nodes()
        if len(self.mdh_backends) > 1:
            if self.args.node_name:
                container = self._lookup_container_by_name(self.args.node_name)[0]
                container.stop()
                container.remove()
                self.show_cluster()
                return
            last_backend_node = self.mdh_backends[-1]
            last_backend_node.stop()
            last_backend_node.remove()
        return self.show_cluster()

    def prepare(self):
        """ Helper to prepare current class

        Returns:
            status (bool): Status of prepare. True on successful info retrieval.
            False if cluster_id is missing from clusters.
        """
        clusters = self._show_all_mdh_cluster_ids()
        if self.cluster_id not in clusters:
            print("--cluster-id is missing")
            return False
        self.cluster_network = self._get_or_create_network()[0]
        self.cluster_data_volume = {self._get_or_create_data_volume()[0].attrs['Name']: {'bind': '/data', 'mode': 'rw'}}
        self.mdh_backends = self.get_all_backend_nodes()
        return True

    def invoke(self):
        """ Helper method for cli
        """

        if "init" in getattr(self.args, "cmd"):
            self.setup_cluster_infra()
        elif "show" in getattr(self.args, "cmd"):
            self.show_cluster()
        elif "delete" in getattr(self.args, "cmd"):
            self.remove_cluster()
        elif "+" in getattr(self.args, "cmd"):
            if self.prepare():
                self.add_node_to_cluster()
        elif "-" in getattr(self.args, "cmd"):
            if self.prepare():
                self.remove_node_from_cluster()

    def remove_cluster(self):
        """ Removes all instances from cluster

        Returns (list): Removed containers statuses

        """
        res = []
        clusters = self._show_all_mdh_cluster_ids()
        if len(clusters) == 1 and not isinstance(self.args, ArgumentParser):
            if clusters[0] == self.cluster_id:
                return res
        for container in self.show_cluster():
            res.append({"stop": container.stop(), "remove": container.remove()})
        return res



import json
import logging.config
import os
import time

import docker
import crossplane

NGINX_CONFIG_FILE = "/data/etc/nginx/nginx.conf"
PROM_TARGET_FILE = "/data/prometheus/targets.json"
TARGETS_PATH = '/'.join(PROM_TARGET_FILE.split("/")[:-1])
DEFAULT_TARGETS = [
    {
        "labels": {
            "job": "mdh_backend"
        },
        "targets": [
            "mdh_backend1:9273"
        ]
    }
]

logging.config.dictConfig({
    'version': 1,
    'formatters': {
        'default': {
            'format': '%(asctime)s [%(levelname)s] %(name)s:%(lineno)d %(funcName)s - %(message)s'
        },
    },
    'handlers': {
        'log': {
            'class': 'logging.StreamHandler',
            'formatter': 'default',
        },
    },
    'loggers': {
        '': {
            'handlers': ['log'],
            'level': 'DEBUG',
        }
    }
})


def write_targets_file(data):
    """ Prometheus targets.json writer

    Args:
        data (object): Object to write JSON file

    """
    with open(PROM_TARGET_FILE, "w") as outfile:
        json.dump(data, outfile)


def load_targets_file():
    """ Load prometheus targets.json file. On missing file new file will be created.

    Returns (object): JSON formatted object

    """
    if not os.path.exists(TARGETS_PATH):
        os.makedirs(TARGETS_PATH)
        write_targets_file(DEFAULT_TARGETS)

    with open(PROM_TARGET_FILE) as f:
        return json.load(f)


def get_all_prom_backends():
    """ Helper function to filter targets from prometheus targets.json file

    Returns (list): List of monitored instances

    """
    return load_targets_file()[0]["targets"]


def add_prom_target(backend):
    """ Adds backend instance for prometheus scraping

    Args:
        backend (str): Backend instance name to add

    Returns:

    """
    backends = list(map(lambda x: x.split(":")[0], get_all_prom_backends()))
    if backend not in backends:
        logging.info("Adding {} to {}".format(backend, PROM_TARGET_FILE))
        targets = load_targets_file()
        targets[0]["targets"].append("{}:9273".format(backend))
        write_targets_file(targets)
        logging.info(targets)


def remove_prom_target(backend):
    """ Removes backend instance from prometheus scraping

    Args:
        backend (str): backend name to remove

    Returns:

    """
    backends = list(map(lambda x: x.split(":")[0], get_all_prom_backends()))
    if backend in backends:
        logging.info("Removinf {} from {}".format(backend, PROM_TARGET_FILE))
        targets = load_targets_file()
        targets[0]["targets"].remove("{}:9273".format(backend))
        write_targets_file(targets)
        logging.info(targets)


def get_all_nginx_backends():
    """ Helper function to filter out from nginx.conf backend servers

    Returns (str): List of backned servers

    """
    parsed_conf = crossplane.parse(NGINX_CONFIG_FILE)
    config_backends = parsed_conf['config'][0]['parsed'][1]['block'][0]['block']
    return list(map(lambda x: x['args'], config_backends))


def add_backend_to_nginx_conf(mdh_backend):
    """ Add backend server to nginx.conf
    Args:
        mdh_backend (dict): Dictionary object  mdh_backend as backend name and mdh_backend_port as backend port

    Returns:

    """
    parsed_conf = crossplane.parse(NGINX_CONFIG_FILE)
    parsed_conf['config'][0]['parsed'][1]['block'][0]['block'].append(
        {'directive': 'server', 'args': ['{mdh_backend}:{mdh_backend_port}'.format(**mdh_backend)]}
    )

    config_str = crossplane.build(parsed_conf['config'][0]['parsed'])
    write_nging_conf(config_str)
    logging.info(config_str)


def write_nging_conf(config_str):
    """ Helper function to weite nginx.file

    Args:
        config_str (str): Raw nginx.conf string

    Returns:

    """

    with open(NGINX_CONFIG_FILE, 'w') as f:
        f.write(config_str)


def remove_backend_from_nginx_conf(mdh_backend):
    """ Remove backend server to nginx.conf
    Args:
        mdh_backend (dict): Dictionary object  where mdh_backend as backend name and mdh_backend_port as backend port

    Returns:

    """
    parsed_conf = crossplane.parse(NGINX_CONFIG_FILE)
    config_backend_names = list(map(lambda x: x[0].split(":")[0], get_all_nginx_backends()))
    logging.info(config_backend_names)
    backend_index = config_backend_names.index('{mdh_backend}'.format(**mdh_backend))
    del parsed_conf['config'][0]['parsed'][1]['block'][0]['block'][backend_index]
    config_str = crossplane.build(parsed_conf['config'][0]['parsed'])
    write_nging_conf(config_str)
    logging.info(config_str)


def handle_nginx_reload(cli, mdh_custer_id):
    lb_container = list(filter(lambda x: (x.labels.get('mdh_lb', "null") == "true" and
                                          x.labels.get('mdh_cluster_id', "null") == mdh_custer_id),
                               cli.containers.list()))[0]
    if lb_container:
        logging.info("Nginx reload")
        lb_container.exec_run("nginx -s reload")
        logging.info(lb_container.logs())


def elsa_docker_event_worker():
    logging.info("Elsa is alive!")
    logging.info("Cluster ID: {}".format(os.environ.get("MDH_CLUSTER_ID")))

    cli = docker.from_env()
    while True:
        try:
            for event in cli.events():
                event = json.loads(event.decode('utf-8'))
                config_backend_names = list(map(lambda x: x[0].split(":")[0], get_all_nginx_backends()))
                if event['Type'] == 'container':
                    mdh_cluster_id = event['Actor']['Attributes'].get('mdh_cluster_id', "no_cluster_id")
                    if os.environ.get("MDH_CLUSTER_ID", "ENV_NOT_SET") != mdh_cluster_id:
                        continue
                    backend_nr = event['Actor']['Attributes'].get('mdh_backend_member_nr')
                    backend_name = "mdh_backend{}".format(backend_nr)
                    if event['Action'] == 'start' and backend_nr:
                        logging.info(event['Actor']['Attributes'])
                        if backend_name in config_backend_names:
                            continue
                        add_backend_to_nginx_conf(dict(mdh_backend=backend_name,
                                                       mdh_backend_port=os.environ.get("MDH_BACKEND_PORT", 5000)))
                        add_prom_target(backend_name)
                        handle_nginx_reload(cli, mdh_cluster_id)
                    elif event['Action'] == 'kill' and backend_nr:
                        remove_backend_from_nginx_conf(dict(mdh_backend=backend_name,
                                                            mdh_backend_port=os.environ.get("MDH_BACKEND_PORT", 5000)))
                        remove_prom_target(backend_name)
                        handle_nginx_reload(cli, mdh_cluster_id)
        except Exception as e:
            logging.exception(e)


def elsa_health_check_worker():
    logging.info("Elsa health worker is  alive!")
    logging.info("Cluster ID: {}".format(os.environ.get("MDH_CLUSTER_ID")))
    statuses = ["exited", "killed"]

    cli = docker.from_env()
    while True:
        mdh_cluster_id = os.environ.get("MDH_CLUSTER_ID", "ENV_NOT_SET")
        containers = cli.containers.list(
            filters={"label": ["mdh_backend=true", "mdh_cluster_id={}".format(mdh_cluster_id)]})
        for container in containers:
            backend_nr = container.labels.get('mdh_backend_member_nr')
            backend_name = "mdh_backend{}".format(backend_nr)
            if container.status in statuses and backend_nr:
                remove_backend_from_nginx_conf(dict(mdh_backend=backend_name,
                                                    mdh_backend_port=os.environ.get("MDH_BACKEND_PORT", 5000)))
                remove_prom_target(backend_name)
        time.sleep(30)



from flask import request
from flask_restful import Resource, reqparse, abort
from .models import Args
from .main import ClusterHandler


class ServiceAPI(Resource):
    def __init__(self):
        self.reqparse = reqparse.RequestParser()
        self.reqparse.add_argument('replicas', type=int, help='Number of cluster members provisioned',
                                   required=True, location='json')
        self.reqparse.add_argument('clusters', type=int, help='Number of clusters provisioned', required=True,
                                   location='json')

        super(ServiceAPI, self).__init__()

    def get(self):
        """ HTTP GET method for /Service api
        Examples:
            /Service?add_service - add's new service
            /Service?add_node&cluster_id=xxxx - add new backend node to cluster
            /Servce?rm_node&cluster_id - removes backend node from cluster
        Returns (list): Cluster status

        """
        res = []
        cluster_id = request.args.get('cluster_id')
        cmd = request.args.get('cmd')
        if cmd == "add_service":
            handler = ClusterHandler(Args(cluster_id=cluster_id))
            handler.prepare()
            res = handler.setup_cluster_infra()
        elif cmd == "add_node" and cluster_id:
            handler = ClusterHandler(Args(cluster_id=cluster_id))
            handler.prepare()
            res = handler.add_node_to_cluster()
        elif cmd == "rm_node" and cluster_id:
            handler = ClusterHandler(Args(cluster_id=cluster_id))
            handler.prepare()
            res = handler.remove_node_from_cluster()
        return res

    def post(self):
        """HTTP POST method for API /Service entrypoint for cluster creation
        Examples:
            POST to /Service entrypoint with data
             { "clusters": "1", "replicas" : "1" }
             creates one cluster with one backned node
        Returns (list): Created cluster info

        """
        posted_data = request.get_json()
        if not posted_data:
            abort("Missing input data")
        replicas = int(posted_data.get("replicas"))
        clusters = int(posted_data.get("clusters"))
        res = []
        if clusters and replicas:
            for cluster in range(clusters):
                args = Args()
                handler = ClusterHandler(args)
                handler.prepare()
                handler.setup_cluster_infra()
                if replicas > 1:
                    handler = ClusterHandler(args)
                    handler.prepare()
                    for replica in range(replicas - 1):
                        handler.add_node_to_cluster()
                res.append({handler.cluster_id: handler.show_cluster()})
        return res

    def patch(self):
        """ HTTP PATCH method for /Service entrypoint for cluster change

        Examples:
            PATCH to /Service?cluster_id=xxx... - With data {"replicas": "1"} changes cluster backend replicas to one


        Returns (list): Updated cluster info

        """
        posted_data = request.get_json()
        if not posted_data:
            abort("Missing input data")
        new_replicas = int(posted_data.get("replicas"))
        cluster_id = request.args.get('cluster_id')
        if not cluster_id:
            abort("Missing cluster_id from args")
        handler = ClusterHandler(Args(cluster_id=cluster_id))
        if handler.prepare():
            return abort("Bad cluster_id")
        if len(handler.mdh_backends) == new_replicas:
            return handler.show_cluster()
        elif len(handler.mdh_backends) > new_replicas:
            while len(handler.mdh_backends) != new_replicas:
                handler.prepare()
                handler.remove_node_from_cluster()
        elif len(handler.mdh_backends) < new_replicas:
            while len(handler.mdh_backends) != new_replicas:
                handler.prepare()
                handler.add_node_to_cluster()
        return handler.show_cluster()

    def delete(self):
        """ HTTP POST method for cluster deletion
        Examples:
            DELETE to /Service?cluster_id=xxx...  - removes cluster

        Returns (list): List of removed containers statuses

        """
        cluster_id = request.args.get('cluster_id')
        if not cluster_id:
            abort("Missing cluster_id from args")
        handler = ClusterHandler(Args(cluster_id=cluster_id))
        if not handler.prepare():
            abort("Bad cluster_id")
        return handler.remove_cluster()


class StateAPI(Resource):
    def get(self):
        """ HTTP GET method for API /State entrypoint
        Examples:
            /State - returns all available clusters info
            /State?cluster_id=xxx.. - returns specified cluster_id info

        Returns (list): Cluster info.

        """
        cluster_id = request.args.get('cluster_id')
        handler = ClusterHandler(Args(cluster_id=cluster_id))
        handler.prepare()
        return handler.show_cluster(show_all=False if cluster_id else True)


class Root(Resource):
    def get(self):
        """ HTTP GET method for API / entrypoint

        Returns (dict): dict formated welcome message

        """
        return {'message': 'welcome'}



"""
MDH backend containers healthcheck worker
"""
import threading
from worker import elsa_docker_event_worker, elsa_health_check_worker
if __name__ == '__main__':
    elsa_docker_worker = threading.Thread(target=elsa_docker_event_worker)
    elsa_docker_worker.start()
    elsa_health_check_worker()
